# encoding=utf-8
import unittest
import pytest
from lxml import etree
from mc2skos.mc2skos import Element


class TestStringify(unittest.TestCase):

    def setUp(self):
        pass

    def testSeeNote(self):
        elem = Element(etree.fromstring(u"""
            <datafield tag="253" ind1="0" ind2=" " xmlns="http://www.loc.gov/MARC21/slim">
                <subfield code="t">Vitenskap og lærdom</subfield>
                <subfield code="i">, se</subfield>
                <subfield code="a">001.2</subfield>
                <subfield code="9">ess=nse</subfield>
            </datafield>
        """))
        assert elem.stringify() == u'Vitenskap og lærdom, se 001.2'

    def testSeeAlsoNote(self):
        elem = Element(etree.fromstring(u"""
            <datafield tag="353" ind1=" " ind2=" " xmlns="http://www.loc.gov/MARC21/slim">
                <subfield code="i">Se også</subfield>
                <subfield code="a">900</subfield>
                <subfield code="i">for en</subfield>
                <subfield code="t">bred beskrivelse av situasjon og vilkår for intellektuell virksomhet</subfield>
                <subfield code="9">ess=nsa</subfield>
            </datafield>
        """))
        assert elem.stringify() == u'Se også 900 for en bred beskrivelse av situasjon og vilkår for intellektuell virksomhet'

    def testNoteWithClassNumberRange(self):
        elem = Element(etree.fromstring(u"""
            <datafield tag="253" ind1="2" ind2=" " xmlns="http://www.loc.gov/MARC21/slim">
                <subfield code="i">Klassifiser</subfield>
                <subfield code="t">andre bestemte internasjonale språk</subfield>
                <subfield code="i">med språket i</subfield>
                <subfield code="a">420</subfield>
                <subfield code="c">490</subfield>
                <subfield code="i">, f.eks.</subfield>
                <subfield code="t">latin som et diplomatspråk</subfield>
                <subfield code="e">470</subfield>
                <subfield code="i">,</subfield>
                <subfield code="t">swahili som et lingua franca</subfield>
                <subfield code="e">496.392</subfield>
                <subfield code="9">ess=ncw</subfield>
            </datafield>
        """))
        assert elem.stringify() == u'Klassifiser andre bestemte internasjonale språk med språket i 420-490, f.eks. latin som et diplomatspråk, swahili som et lingua franca'

    def testComplexNote(self):
        elem = Element(etree.fromstring(u"""
            <datafield tag="680" ind1="0" ind2=" " xmlns="http://www.loc.gov/MARC21/slim">
                <subfield code="i">Inkluderer:</subfield>
                <subfield code="t">Case-studier</subfield>
                <subfield code="i">[tidligere</subfield>
                <subfield code="x">001.432</subfield>
                <subfield code="i">];</subfield>
                <subfield code="t">utvalgsteknikker</subfield>
                <subfield code="i">;</subfield>
                <subfield code="t">rundspørringer</subfield>
                <subfield code="i">,</subfield>
                <subfield code="t">spørreskjemaer</subfield>
                <subfield code="i">,</subfield>
                <subfield code="t">feltarbeid</subfield>
                <subfield code="i">,</subfield>
                <subfield code="t">deltakende observasjon</subfield>
                <subfield code="i">,</subfield>
                <subfield code="t">intervjuer</subfield>
                <subfield code="9">ess=nin</subfield>
            </datafield>
        """))
        assert elem.stringify() == u'Inkluderer: Case-studier [tidligere 001.432]; utvalgsteknikker; rundspørringer, spørreskjemaer, feltarbeid, deltakende observasjon, intervjuer'


if __name__ == '__main__':
    unittest.main()



# encoding=utf-8
import unittest
import pytest
from rdflib import Graph, URIRef, Namespace
from rdflib.namespace import SKOS

from mc2skos.mc2skos import process_records
from mc2skos.reader import MarcFileReader
from mc2skos.vocabularies import Vocabularies

BK = Namespace('http://uri.gbv.de/terminology/bk/')


with open('mc2skos/vocabularies.yml') as fp:
    vocabularies = Vocabularies()
    vocabularies.load_yaml(fp)


def test_skosify():
    marc = MarcFileReader('examples/bk-54.65.xml')
    rdf = process_records(marc.records(), expand=True, vocabularies=vocabularies)

    assert (BK['54'], SKOS.narrower, BK['54.65']) in rdf



# encoding=utf-8
import unittest
import pytest
from lxml import etree
from mc2skos.mc2skos import ClassificationRecord, Element


class TestParse153(unittest.TestCase):

    def testSimpleClass(self):
        element = Element('''
            <marc:datafield tag="153" ind1=" " ind2=" " xmlns:marc="http://www.loc.gov/MARC21/slim">
              <marc:subfield code="a">003.5</marc:subfield>
              <marc:subfield code="e">003</marc:subfield>
              <marc:subfield code="h">Generalities</marc:subfield>
              <marc:subfield code="h">Systems</marc:subfield>
              <marc:subfield code="j">Theory of communication and control</marc:subfield>
            </marc:datafield>
        ''')

        table, notation, is_top_concept, parent_notation, caption = ClassificationRecord.parse_153(element)

        assert notation == '003.5'
        assert parent_notation == '003'
        assert is_top_concept is False
        assert caption == 'Theory of communication and control'

    def testTableAddTableEntry(self):
        element = Element('''
            <mx:datafield tag="153" ind2=" " ind1=" " xmlns:mx="http://www.loc.gov/MARC21/slim">
                <mx:subfield code="z">3B</mx:subfield>
                <mx:subfield code="a">81</mx:subfield>
                <mx:subfield code="c">89</mx:subfield>
                <mx:subfield code="y">1</mx:subfield>
                <mx:subfield code="a">02</mx:subfield>
                <mx:subfield code="z">3B</mx:subfield>
                <mx:subfield code="e">81</mx:subfield>
                <mx:subfield code="f">89</mx:subfield>
                <mx:subfield code="j">Anekdoter, epigrammer, graffiti, vitser, vittigheter, sitater, gåter, tungekrøllere</mx:subfield>
                <mx:subfield code="9">ess=ren</mx:subfield>
                <mx:subfield code="9">ess=reh</mx:subfield>
            </mx:datafield>

        ''')

        table, notation, is_top_concept, parent_notation, caption = ClassificationRecord.parse_153(element)

        assert table == '3B'
        assert notation == '3B--81-89:02'
        assert is_top_concept is False
        assert parent_notation == '3B--81-89'
        assert caption == u'Anekdoter, epigrammer, graffiti, vitser, vittigheter, sitater, gåter, tungekrøllere'

    def testAddTableEntry(self):
        element = Element('''
            <mx:datafield tag="153" ind2=" " ind1=" " xmlns:mx="http://www.loc.gov/MARC21/slim">
                <mx:subfield code="a">820.1</mx:subfield>
                <mx:subfield code="c">828</mx:subfield>
                <mx:subfield code="y">4</mx:subfield>
                <mx:subfield code="a">1</mx:subfield>
                <mx:subfield code="e">820</mx:subfield>
                <mx:subfield code="j">Early period to 1858</mx:subfield>
                <mx:subfield code="9">ess=reb</mx:subfield>
                <mx:subfield code="9">ess=rhb</mx:subfield>
            </mx:datafield>
        ''')

        table, notation, is_top_concept, parent_notation, caption = ClassificationRecord.parse_153(element)

        assert table is None
        assert notation == '820.1-828:4;1'
        assert is_top_concept is False
        assert parent_notation == '820'
        assert caption == u'Early period to 1858'

    def testTableEntryOldStyle(self):
        element = Element('''
            <marc:datafield tag="153" ind1=" " ind2=" " xmlns:marc="http://www.loc.gov/MARC21/slim">
                <marc:subfield code="z">6</marc:subfield>
                <marc:subfield code="a">9839</marc:subfield>
                <marc:subfield code="h">Languages</marc:subfield>
                <marc:subfield code="h">Other languages</marc:subfield>
                <marc:subfield code="h">South American native languages</marc:subfield>
                <marc:subfield code="h">Quechuan (Kechuan), Aymaran, Tucanoan, Tupí, Arawakan languages</marc:subfield>
                <marc:subfield code="j">Arawakan languages</marc:subfield>
            </marc:datafield>
        ''')

        table, notation, is_top_concept, parent_notation, caption = ClassificationRecord.parse_153(element)

        assert table == '6'
        assert notation == '6--9839'
        assert is_top_concept is False
        assert parent_notation is None
        assert caption == u'Arawakan languages'

    def testComplexTableEntryWithUndocumentStuff(self):
        # Test that none of the extra stuff (after $f) leaks into the notation
        element = Element('''
            <mx:datafield tag="153" ind2=" " ind1=" " xmlns:mx="http://www.loc.gov/MARC21/slim">
                <mx:subfield code="z">1</mx:subfield>
                <mx:subfield code="a">0926</mx:subfield>
                <mx:subfield code="z">1</mx:subfield>
                <mx:subfield code="e">0923</mx:subfield>
                <mx:subfield code="f">0928</mx:subfield>
                <mx:subfield code="j">Samlingsbiografier om personer inndelt etter diverse sosiale kjennetegn</mx:subfield>
                <mx:subfield code="i">[tidligere</mx:subfield>
                <mx:subfield code="z">1</mx:subfield>
                <mx:subfield code="x">0922</mx:subfield>
                <mx:subfield code="i">,</mx:subfield>
                <mx:subfield code="z">1</mx:subfield>
                <mx:subfield code="x">0923</mx:subfield>
                <mx:subfield code="i">]</mx:subfield>
                <mx:subfield code="9">ess=ten</mx:subfield>
                <mx:subfield code="9">ess=eh</mx:subfield>
                <mx:subfield code="9">ess=nrl</mx:subfield>
              </mx:datafield>
        ''')

        table, notation, is_top_concept, parent_notation, caption = ClassificationRecord.parse_153(element)

        assert table == '1'
        assert notation == '1--0926'
        assert is_top_concept is False
        assert parent_notation == '1--0923-0928'
        assert caption == u'Samlingsbiografier om personer inndelt etter diverse sosiale kjennetegn'

    def testStandardSubdivisionInfo(self):
        element = Element('''
            <mx:datafield tag="153" ind2=" " ind1=" " xmlns:mx="http://www.loc.gov/MARC21/slim">
                <mx:subfield code="a">973</mx:subfield>
                <mx:subfield code="9">ess=si1</mx:subfield>
            </mx:datafield>
        ''')

        table, notation, is_top_concept, parent_notation, caption = ClassificationRecord.parse_153(element)

        assert table is None
        assert notation == '973'
        assert caption is None

    def testSynthesizedNumber(self):
        element = Element('''
              <mx:datafield tag="153" ind2=" " ind1=" " xmlns:mx="http://www.loc.gov/MARC21/slim">
                <mx:subfield code="a">001.4092</mx:subfield>
                <mx:subfield code="e">001.4</mx:subfield>
                <mx:subfield code="9">ess=ien</mx:subfield>
              </mx:datafield>
        ''')

        table, notation, is_top_concept, parent_notation, caption = ClassificationRecord.parse_153(element)

        assert table is None
        assert is_top_concept is False
        assert notation == '001.4092'
        assert parent_notation == '001.4'
        assert caption is None

    def testExtraSubfields(self):
        element = Element('''
              <mx:datafield tag="153" ind2=" " ind1=" " xmlns:mx="http://www.loc.gov/MARC21/slim">
                <mx:subfield code="a">332.0240081</mx:subfield>
                <mx:subfield code="c">332.0240088</mx:subfield>
                <mx:subfield code="e">332.024001</mx:subfield>
                <mx:subfield code="f">332.024009</mx:subfield>
                <mx:subfield code="j">Miscellaneous specific kinds of persons</mx:subfield>
                <mx:subfield code="i">[formerly</mx:subfield>
                <mx:subfield code="x">332.02404</mx:subfield>
                <mx:subfield code="c">332.0249</mx:subfield>
                <mx:subfield code="i">]</mx:subfield>
                <mx:subfield code="9">ess=en</mx:subfield>
                <mx:subfield code="9">ess=eh</mx:subfield>
                <mx:subfield code="9">ess=nrl</mx:subfield>
              </mx:datafield>
        ''')

        table, notation, is_top_concept, parent_notation, caption = ClassificationRecord.parse_153(element)

        assert table is None
        assert is_top_concept is False
        assert notation == '332.0240081-332.0240088'
        assert parent_notation == '332.024001-332.024009'
        assert caption == 'Miscellaneous specific kinds of persons'


if __name__ == '__main__':
    unittest.main()



# encoding=utf-8
import copy
import unittest
import pytest
from lxml import etree
from mc2skos.mc2skos import process_record, ClassificationRecord, Constants, InvalidRecordError
from mc2skos.vocabularies import Vocabularies
from rdflib.namespace import RDF, SKOS, Namespace
from rdflib import URIRef, Literal, Graph, BNode


with open('mc2skos/vocabularies.yml') as fp:
    vocabularies = Vocabularies()
    vocabularies.load_yaml(fp)


class TestClassificationRecord(unittest.TestCase):

    def setUp(self):
        self.options = {'vocabularies': vocabularies}

    def testSimpleNumber(self):
        rec = ClassificationRecord('''
            <mx:record xmlns:mx="http://www.loc.gov/MARC21/slim">
              <mx:leader>00000nw  a2200000n  4500</mx:leader>
              <mx:controlfield tag="008">091203aaaaaaaa</mx:controlfield>
              <mx:datafield tag="084" ind2=" " ind1="0">
                <mx:subfield code="a">ddc</mx:subfield>
                <mx:subfield code="c">23no</mx:subfield>
                <mx:subfield code="e">nob</mx:subfield>
              </mx:datafield>
              <mx:datafield tag="153" ind2=" " ind1=" ">
                <mx:subfield code="a">152</mx:subfield>
                <mx:subfield code="e">152</mx:subfield>
                <mx:subfield code="f">158</mx:subfield>
                <mx:subfield code="j">Sansing, bevegelse, emosjoner, fysiologiske drifter</mx:subfield>
                <mx:subfield code="9">ess=en</mx:subfield>
                <mx:subfield code="9">ess=eh</mx:subfield>
              </mx:datafield>
            </mx:record>
        ''', options=self.options)

        assert rec.record_type == Constants.SCHEDULE_RECORD
        assert rec.number_type == Constants.SINGLE_NUMBER
        assert rec.display is True
        assert rec.synthesized is False

    def testAddTableNumber(self):
        rec = ClassificationRecord('''
            <mx:record xmlns:mx="http://www.loc.gov/MARC21/slim">
                <mx:leader>00000nw  a2200000n  4500</mx:leader>
                <mx:controlfield tag="008">100414baabaaaa</mx:controlfield>
                <mx:datafield tag="084" ind2=" " ind1="0">
                    <mx:subfield code="a">ddc</mx:subfield>
                    <mx:subfield code="c">23no</mx:subfield>
                </mx:datafield>
                <mx:datafield tag="153" ind2=" " ind1=" ">
                    <mx:subfield code="a">811</mx:subfield>
                    <mx:subfield code="c">818</mx:subfield>
                    <mx:subfield code="y">2</mx:subfield>
                    <mx:subfield code="a">4</mx:subfield>
                    <mx:subfield code="e">811</mx:subfield>
                    <mx:subfield code="f">818</mx:subfield>
                    <mx:subfield code="j">1867-1899 i Canada</mx:subfield>
                    <mx:subfield code="9">ess=reb</mx:subfield>
                    <mx:subfield code="9">ess=rhb</mx:subfield>
                </mx:datafield>
            </mx:record>
        ''', options=self.options)

        assert rec.record_type == Constants.TABLE_RECORD
        assert rec.number_type == Constants.SINGLE_NUMBER
        assert rec.display is True
        assert rec.synthesized is False
        assert rec.notation == '811-818:2;4'
        assert len(rec.relations) == 1
        assert rec.relations[0]['uri'] == 'http://dewey.info/class/811-818/e23/'
        assert rec.relations[0]['relation'] == SKOS.broader

    def testHistoricalAddTableNumber(self):
        rec = ClassificationRecord('''
        <mx:record xmlns:mx="http://www.loc.gov/MARC21/slim">
            <mx:leader>00000nw  a2200000n  4500</mx:leader>
            <mx:controlfield tag="008">091203baaaaaah</mx:controlfield>
            <mx:datafield tag="084" ind2=" " ind1="0">
                <mx:subfield code="a">ddc</mx:subfield>
                <mx:subfield code="c">23no</mx:subfield>
                <mx:subfield code="e">nob</mx:subfield>
            </mx:datafield>
            <mx:datafield tag="153" ind2=" " ind1=" ">
                <mx:subfield code="a">820.1</mx:subfield>
                <mx:subfield code="c">828</mx:subfield>
                <mx:subfield code="y">1</mx:subfield>
                <mx:subfield code="a">7</mx:subfield>
                <mx:subfield code="e">820</mx:subfield>
                <mx:subfield code="j">1800-1837</mx:subfield>
                <mx:subfield code="9">ess=reb</mx:subfield>
                <mx:subfield code="9">ess=rhb</mx:subfield>
            </mx:datafield>
        </mx:record>
        ''', options=self.options)

        assert rec.record_type == Constants.TABLE_RECORD
        assert rec.number_type == Constants.SINGLE_NUMBER
        assert rec.display is False
        assert rec.synthesized is False

    def testSynthesizedNumberSpan(self):
        rec = ClassificationRecord('''
        <mx:record xmlns:mx="http://www.loc.gov/MARC21/slim">
          <mx:leader>00000nw  a2200000n  4500</mx:leader>
          <mx:controlfield tag="008">091203abdaaaba</mx:controlfield>
          <mx:datafield tag="084" ind2=" " ind1="0">
            <mx:subfield code="a">ddc</mx:subfield>
            <mx:subfield code="c">23no</mx:subfield>
            <mx:subfield code="e">nob</mx:subfield>
          </mx:datafield>
          <mx:datafield tag="153" ind2=" " ind1=" ">
            <mx:subfield code="a">133.01</mx:subfield>
            <mx:subfield code="c">133.09</mx:subfield>
            <mx:subfield code="e">133</mx:subfield>
            <mx:subfield code="j">Generell forminndeling</mx:subfield>
            <mx:subfield code="9">ess=en</mx:subfield>
            <mx:subfield code="9">ess=eh</mx:subfield>
          </mx:datafield>
        </mx:record>
        ''', options=self.options)

        assert rec.record_type == Constants.SCHEDULE_RECORD
        assert rec.number_type == Constants.NUMBER_SPAN
        assert rec.display is True
        assert rec.synthesized is True

    def testSynthesizedScheduleRecord(self):
        rec = ClassificationRecord('''
        <mx:record xmlns:mx="http://www.loc.gov/MARC21/slim">
          <mx:leader>00000nw  a2200000n  4500</mx:leader>
          <mx:controlfield tag="008">091203aaaaaabb</mx:controlfield>
          <mx:datafield tag="084" ind2=" " ind1="0">
            <mx:subfield code="a">ddc</mx:subfield>
            <mx:subfield code="c">23no</mx:subfield>
            <mx:subfield code="e">nob</mx:subfield>
          </mx:datafield>
          <mx:datafield tag="153" ind2=" " ind1=" ">
            <mx:subfield code="a">025.1712</mx:subfield>
            <mx:subfield code="e">025.17</mx:subfield>
            <mx:subfield code="9">ess=ien</mx:subfield>
          </mx:datafield>
        </mx:record>
        ''', options=self.options)

        assert rec.record_type == Constants.SCHEDULE_RECORD
        assert rec.number_type == Constants.SINGLE_NUMBER
        assert rec.display is True
        assert rec.synthesized is True

    def testSynthesizedRecordWithIndexTerms(self):
        rec = ClassificationRecord('''
          <mx:record xmlns:mx="http://www.loc.gov/MARC21/slim">
            <mx:leader>00000nw  a2200000n  4500</mx:leader>
            <mx:controlfield tag="008">091203aaaaaabb</mx:controlfield>
            <mx:datafield tag="084" ind2=" " ind1="0">
              <mx:subfield code="a">ddc</mx:subfield>
              <mx:subfield code="c">23no</mx:subfield>
              <mx:subfield code="e">nob</mx:subfield>
            </mx:datafield>
            <mx:datafield tag="153" ind2=" " ind1=" ">
              <mx:subfield code="a">006.686</mx:subfield>
              <mx:subfield code="e">006.68</mx:subfield>
              <mx:subfield code="9">ess=ien</mx:subfield>
            </mx:datafield>
            <mx:datafield tag="730" ind2="7" ind1=" ">
              <mx:subfield code="a">CorelDRAW</mx:subfield>
              <mx:subfield code="0">(OCoLC-D)99928acf-5fcb-4b49-b12c-9d4bdcb369b1</mx:subfield>
              <mx:subfield code="2">ddcri</mx:subfield>
              <mx:subfield code="9">ps=EO</mx:subfield>
            </mx:datafield>
            <mx:datafield tag="750" ind2="7" ind1=" ">
              <mx:subfield code="a">CorelDRAW!</mx:subfield>
              <mx:subfield code="0">(OCoLC-D)73e62e00-d820-48bb-b5a6-d5f4ceab2d12</mx:subfield>
              <mx:subfield code="2">ddcri</mx:subfield>
              <mx:subfield code="9">ps=EO</mx:subfield>
            </mx:datafield>
            <mx:datafield tag="750" ind2="7" ind1=" ">
              <mx:subfield code="a">Personlige datamaskiner</mx:subfield>
              <mx:subfield code="x">grafikkprogrammer</mx:subfield>
              <mx:subfield code="0">(OCoLC-D)1226b03f-c205-420e-ae21-34d41be81715</mx:subfield>
              <mx:subfield code="2">ddcri</mx:subfield>
              <mx:subfield code="9">ps=PE</mx:subfield>
              <mx:subfield code="9">ess=isCaption</mx:subfield>
            </mx:datafield>
          </mx:record>
        ''', options=self.options)

        assert rec.record_type == Constants.SCHEDULE_RECORD
        assert rec.number_type == Constants.SINGLE_NUMBER
        assert rec.display is True
        assert rec.synthesized is True
        assert rec.prefLabel is None
        assert rec.altLabel == [
            {'term': 'Personlige datamaskiner--grafikkprogrammer'},
            {'term': 'CorelDRAW'},
            {'term': 'CorelDRAW!'},
        ]

    def testSynthesizedNumberComponents1(self):
        rec = ClassificationRecord('''
        <mx:record xmlns:mx="http://www.loc.gov/MARC21/slim">
          <mx:leader>00000nw  a2200000n  4500</mx:leader>
          <mx:controlfield tag="001">ocd00132963</mx:controlfield>
          <mx:controlfield tag="008">100204aaaaaabb</mx:controlfield>
          <mx:datafield tag="084" ind2=" " ind1="0">
            <mx:subfield code="a">ddc</mx:subfield>
            <mx:subfield code="c">23no</mx:subfield>
            <mx:subfield code="e">nob</mx:subfield>
          </mx:datafield>
          <mx:datafield tag="153" ind2=" " ind1=" ">
            <mx:subfield code="a">306.6804</mx:subfield>
            <mx:subfield code="e">306.63</mx:subfield>
            <mx:subfield code="f">306.69</mx:subfield>
            <mx:subfield code="9">ess=ien</mx:subfield>
          </mx:datafield>
          <mx:datafield tag="765" ind2=" " ind1="0">
              <mx:subfield code="b">306.6</mx:subfield>
              <mx:subfield code="a">306.63</mx:subfield>
              <mx:subfield code="c">306.69</mx:subfield>
              <mx:subfield code="r">2</mx:subfield>
              <mx:subfield code="s">804</mx:subfield>
              <mx:subfield code="u">306.6804</mx:subfield>
              <mx:subfield code="9">ess=hn</mx:subfield>
            </mx:datafield>
        </mx:record>
        ''', options=self.options)

        assert rec.components == ['306.6', '280.4']

    def testSynthesizedNumberComponents2(self):
        rec = ClassificationRecord('''
        <mx:record xmlns:mx="http://www.loc.gov/MARC21/slim">
          <mx:leader>00000nw  a2200000n  4500</mx:leader>
          <mx:controlfield tag="001">ocd00123528</mx:controlfield>
          <mx:controlfield tag="008">091203aaaaaabb</mx:controlfield>
          <mx:datafield tag="084" ind2=" " ind1="0">
            <mx:subfield code="a">ddc</mx:subfield>
            <mx:subfield code="c">23no</mx:subfield>
            <mx:subfield code="e">nob</mx:subfield>
          </mx:datafield>
          <mx:datafield tag="153" ind2=" " ind1=" ">
            <mx:subfield code="a">299.3113</mx:subfield>
            <mx:subfield code="e">299.31</mx:subfield>
            <mx:subfield code="9">ess=ien</mx:subfield>
          </mx:datafield>
          <mx:datafield tag="765" ind2=" " ind1="0">
              <mx:subfield code="b">299.31</mx:subfield>
              <mx:subfield code="a">299.31</mx:subfield>
              <mx:subfield code="a">290</mx:subfield>
              <mx:subfield code="w">290</mx:subfield>
              <mx:subfield code="y">1</mx:subfield>
              <mx:subfield code="a">1</mx:subfield>
              <mx:subfield code="c">9</mx:subfield>
              <mx:subfield code="r">20</mx:subfield>
              <mx:subfield code="s">13</mx:subfield>
              <mx:subfield code="u">299.3113</mx:subfield>
            </mx:datafield>
            <mx:datafield tag="765" ind2=" " ind1="0">
              <mx:subfield code="b">299</mx:subfield>
              <mx:subfield code="a">299.1</mx:subfield>
              <mx:subfield code="c">299.4</mx:subfield>
              <mx:subfield code="z">5</mx:subfield>
              <mx:subfield code="r">9</mx:subfield>
              <mx:subfield code="s">31</mx:subfield>
              <mx:subfield code="u">299.31</mx:subfield>
            </mx:datafield>
        </mx:record>
        ''', options=self.options)

        assert rec.components == ['299', '5--931', '201.3']

    @pytest.mark.skip(reason="add table numbers not yet supported")
    def testSynthesizedNumberComponentsIncludingAddTable(self):
        rec = ClassificationRecord('''
        <mx:record xmlns:mx="http://www.loc.gov/MARC21/slim">
          <mx:leader>00000nw  a2200000n  4500</mx:leader>
          <mx:controlfield tag="001">ocd00117858</mx:controlfield>
          <mx:controlfield tag="003">OCoLC-D</mx:controlfield>
          <mx:controlfield tag="005">20150910004647.0</mx:controlfield>
          <mx:controlfield tag="008">091203aaaaaabb</mx:controlfield>
          <mx:datafield tag="040" ind2=" " ind1=" ">
            <mx:subfield code="a">OCLCD</mx:subfield>
            <mx:subfield code="b">nob</mx:subfield>
            <mx:subfield code="c">OCLCD</mx:subfield>
          </mx:datafield>
          <mx:datafield tag="084" ind2=" " ind1="0">
            <mx:subfield code="a">ddc</mx:subfield>
            <mx:subfield code="c">23no</mx:subfield>
            <mx:subfield code="e">nob</mx:subfield>
          </mx:datafield>
          <mx:datafield tag="153" ind2=" " ind1=" ">
            <mx:subfield code="a">032.020993</mx:subfield>
            <mx:subfield code="e">032.02</mx:subfield>
            <mx:subfield code="9">ess=ien</mx:subfield>
          </mx:datafield>
          <mx:datafield tag="765" ind2=" " ind1="0">
            <mx:subfield code="b">032.0209</mx:subfield>
            <mx:subfield code="z">1</mx:subfield>
            <mx:subfield code="a">093</mx:subfield>
            <mx:subfield code="c">099</mx:subfield>
            <mx:subfield code="z">2</mx:subfield>
            <mx:subfield code="s">93</mx:subfield>
            <mx:subfield code="u">032.020993</mx:subfield>
          </mx:datafield>
          <mx:datafield tag="765" ind2=" " ind1="0">
            <mx:subfield code="b">032.02</mx:subfield>
            <mx:subfield code="z">1</mx:subfield>
            <mx:subfield code="s">09</mx:subfield>
            <mx:subfield code="u">032.0209</mx:subfield>
          </mx:datafield>
          <mx:datafield tag="765" ind2=" " ind1="0">
            <mx:subfield code="b">032</mx:subfield>
            <mx:subfield code="a">032</mx:subfield>
            <mx:subfield code="a">031</mx:subfield>
            <mx:subfield code="c">039</mx:subfield>
            <mx:subfield code="w">031</mx:subfield>
            <mx:subfield code="c">039</mx:subfield>
            <mx:subfield code="y">1</mx:subfield>
            <mx:subfield code="t">02</mx:subfield>
            <mx:subfield code="u">032.02</mx:subfield>
          </mx:datafield>
          <mx:datafield tag="750" ind2="7" ind1=" ">
            <mx:subfield code="a">Engelske almanakker</mx:subfield>
            <mx:subfield code="z">New Zealand</mx:subfield>
            <mx:subfield code="0">(OCoLC-D)792f96bb-142c-43c6-a20e-ed5ed2088388</mx:subfield>
            <mx:subfield code="2">ddcri</mx:subfield>
            <mx:subfield code="9">ps=EO</mx:subfield>
          </mx:datafield>
        </mx:record>
        ''', options=self.options)

        assert rec.components == ['032', '031-039:02', '1--09', '2--93']

    def testIndexTerms(self):
        rec = ClassificationRecord('''
        <mx:record xmlns:mx="http://www.loc.gov/MARC21/slim">
          <mx:leader>00000nw  a2200000n  4500</mx:leader>
          <mx:controlfield tag="001">ocd00146759</mx:controlfield>
          <mx:controlfield tag="008">100204aaaaaaaa</mx:controlfield>
          <mx:datafield tag="084" ind2=" " ind1="0">
            <mx:subfield code="a">ddc</mx:subfield>
            <mx:subfield code="c">23no</mx:subfield>
            <mx:subfield code="e">nob</mx:subfield>
          </mx:datafield>
          <mx:datafield tag="153" ind2=" " ind1=" ">
            <mx:subfield code="a">543.17</mx:subfield>
            <mx:subfield code="e">543.1</mx:subfield>
            <mx:subfield code="j">Analytisk organisk kjemi</mx:subfield>
            <mx:subfield code="9">ess=en</mx:subfield>
            <mx:subfield code="9">ess=eh</mx:subfield>
          </mx:datafield>
          <mx:datafield tag="750" ind2="7" ind1=" ">
            <mx:subfield code="a">Analytisk kjemi</mx:subfield>
            <mx:subfield code="x">organisk kjemi</mx:subfield>
            <mx:subfield code="0">(OCoLC-D)8c2057ce-4544-4593-9699-1008a7dcd4ef</mx:subfield>
            <mx:subfield code="2">ddcri</mx:subfield>
            <mx:subfield code="9">ps=PE</mx:subfield>
          </mx:datafield>
          <mx:datafield tag="750" ind2="7" ind1=" ">
            <mx:subfield code="a">Kjemisk analyse</mx:subfield>
            <mx:subfield code="x">organisk kjemi</mx:subfield>
            <mx:subfield code="0">(OCoLC-D)934d9916-e069-4351-994f-44b4c02f2f4d</mx:subfield>
            <mx:subfield code="2">ddcri</mx:subfield>
            <mx:subfield code="9">ps=PE</mx:subfield>
          </mx:datafield>
          <mx:datafield tag="750" ind2="7" ind1=" ">
            <mx:subfield code="a">Organisk kjemi</mx:subfield>
            <mx:subfield code="x">analytisk kjemi</mx:subfield>
            <mx:subfield code="0">(OCoLC-D)9d92b5a2-7f96-4db7-a212-19c40edf7a93</mx:subfield>
            <mx:subfield code="2">ddcri</mx:subfield>
            <mx:subfield code="9">ps=PE</mx:subfield>
          </mx:datafield>
        </mx:record>
        ''', options=self.options)

        assert rec.altLabel == [
            {'term': 'Analytisk kjemi--organisk kjemi'},
            {'term': 'Kjemisk analyse--organisk kjemi'},
            {'term': 'Organisk kjemi--analytisk kjemi'}]


class TestProcessRecord(unittest.TestCase):

    def setUp(self):
        self.graph = Graph()
        self.nsmap = {'mx': 'http://www.loc.gov/MARC21/slim'}
        self.vocabularies = copy.deepcopy(vocabularies)

    def testEmptyRecord(self):
        g = Graph()

        rec = '''
          <marc:record xmlns:marc="http://www.loc.gov/MARC21/slim">
          </marc:record>
        '''

        with pytest.raises(InvalidRecordError):
            process_record(g, rec, vocabularies=self.vocabularies)

    def testRecordWithInvalidLeader(self):

        # A Marc21 Bibliographic record
        rec = '''
          <marc:record xmlns:marc="http://www.loc.gov/MARC21/slim">
            <marc:leader>00000aa  a2200000n  4500</marc:leader>
            <marc:datafield tag="084" ind2=" " ind1="0">
              <marc:subfield code="a">ddc</marc:subfield>
              <marc:subfield code="c">23no</marc:subfield>
              <marc:subfield code="e">nob</marc:subfield>
            </marc:datafield>
          </marc:record>
        '''

        with pytest.raises(InvalidRecordError):
            process_record(Graph(), rec, vocabularies=self.vocabularies)

    def testRecordWithout153(self):
        # Not sure if this test should fail or not.

        rec = '''
          <marc:record xmlns:marc="http://www.loc.gov/MARC21/slim">
            <marc:leader>00000nw  a2200000n  4500</marc:leader>
            <marc:datafield tag="084" ind2=" " ind1="0">
              <marc:subfield code="a">ddc</marc:subfield>
              <marc:subfield code="c">23no</marc:subfield>
              <marc:subfield code="e">nob</marc:subfield>
            </marc:datafield>
          </marc:record>
        '''

        with pytest.raises(InvalidRecordError):
            process_record(Graph(), rec, vocabularies=self.vocabularies)

    def test153(self):
        rec = '''
          <marc:record xmlns:marc="http://www.loc.gov/MARC21/slim">
            <marc:leader>00000nw  a2200000n  4500</marc:leader>
            <marc:datafield tag="153" ind1=" " ind2=" ">
              <marc:subfield code="a">003.5</marc:subfield>
              <marc:subfield code="e">003</marc:subfield>
              <marc:subfield code="h">Generalities</marc:subfield>
              <marc:subfield code="h">Systems</marc:subfield>
              <marc:subfield code="j">Theory of communication and control</marc:subfield>
            </marc:datafield>
          </marc:record>
        '''
        graph = Graph()
        self.vocabularies.set_default_scheme('http://test/{object}')
        process_record(graph, rec, vocabularies=self.vocabularies)
        uri = URIRef(u'http://test/003.5')

        assert set(graph) == set([
            (uri, RDF.type, SKOS.Concept),
            (uri, SKOS.broader, URIRef(u'http://test/003')),
            (uri, SKOS.prefLabel, Literal('Theory of communication and control', lang='en')),
            (uri, SKOS.inScheme, URIRef(u'http://test/')),
            (uri, SKOS.notation, Literal('003.5')),
        ])

    def test_language(self):
        rec = '''
          <marc:record xmlns:marc="http://www.loc.gov/MARC21/slim">
            <marc:leader>00000nw  a2200000n  4500</marc:leader>
            <marc:datafield tag="040" ind2=" " ind1=" ">
              <marc:subfield code="a">OCLCD</marc:subfield>
              <marc:subfield code="b">nob</marc:subfield>
              <marc:subfield code="c">OCLCD</marc:subfield>
            </marc:datafield>
            <marc:datafield tag="084" ind2=" " ind1="0">
              <marc:subfield code="a">ddc</marc:subfield>
              <marc:subfield code="c">23no</marc:subfield>
              <marc:subfield code="e">nob</marc:subfield>
            </marc:datafield>
            <marc:datafield tag="153" ind2=" " ind1=" ">
              <marc:subfield code="a">564.58</marc:subfield>
              <marc:subfield code="e">564.5</marc:subfield>
              <marc:subfield code="j">Decapoda (tiarmede blekkspruter)</marc:subfield>
            </marc:datafield>
          </marc:record>
        '''
        graph = Graph()
        self.vocabularies.set_default_scheme('http://test/{object}')
        process_record(graph, rec, vocabularies=self.vocabularies)
        uri = URIRef(u'http://test/564.58')

        assert graph.preferredLabel(uri)[0][0] == SKOS.prefLabel
        assert graph.preferredLabel(uri)[0][1].language == 'nb'

    def testSynthesizedNumberComponents(self):
        rec = '''
        <mx:record xmlns:mx="http://www.loc.gov/MARC21/slim">
          <mx:leader>00000nw  a2200000n  4500</mx:leader>
          <mx:controlfield tag="001">ocd00132963</mx:controlfield>
          <mx:controlfield tag="008">100204aaaaaabb</mx:controlfield>
          <mx:datafield tag="084" ind2=" " ind1="0">
            <mx:subfield code="a">ddc</mx:subfield>
            <mx:subfield code="c">23no</mx:subfield>
          </mx:datafield>
          <mx:datafield tag="153" ind2=" " ind1=" ">
            <mx:subfield code="a">306.6804</mx:subfield>
            <mx:subfield code="e">306.63</mx:subfield>
            <mx:subfield code="f">306.69</mx:subfield>
            <mx:subfield code="9">ess=ien</mx:subfield>
          </mx:datafield>
          <mx:datafield tag="765" ind2=" " ind1="0">
            <mx:subfield code="b">306.6</mx:subfield>
            <mx:subfield code="a">306.63</mx:subfield>
            <mx:subfield code="c">306.69</mx:subfield>
            <mx:subfield code="r">2</mx:subfield>
            <mx:subfield code="s">804</mx:subfield>
            <mx:subfield code="u">306.6804</mx:subfield>
            <mx:subfield code="9">ess=hn</mx:subfield>
          </mx:datafield>
        </mx:record>
        '''

        graph = Graph()
        process_record(graph, rec, include_components=True, vocabularies=self.vocabularies)

        components = [n[0] for n in graph.query('''
            PREFIX mads: <http://www.loc.gov/mads/rdf/v1#>
            SELECT ?comp WHERE {
              <http://dewey.info/class/306.6804/e23/> mads:componentList/rdf:rest*/rdf:first ?comp
            }
        ''')]

        assert len(components) == 2
        assert components == [URIRef(u'http://dewey.info/class/306.6/e23/'),
                              URIRef(u'http://dewey.info/class/280.4/e23/')]


if __name__ == '__main__':
    unittest.main()



def is_uri(value):
    return value.startswith('http://') or value.startswith('https://')


def is_str(obj):
    try:
        return isinstance(obj, basestring)  # Python 2.x
    except NameError:
        return isinstance(obj, str)  # Python 3.x



#!/usr/bin/env python
# encoding=utf8
#
# Script to convert MARC 21 Classification records
# (serialized as MARCXML) to SKOS concepts. See
# README.md for more information.

import sys
import re
import time
import warnings
from datetime import datetime
from iso639 import languages
import argparse
from rdflib.namespace import OWL, RDF, SKOS, DCTERMS, XSD, Namespace
from rdflib import URIRef, Literal, Graph, BNode
from otsrdflib import OrderedTurtleSerializer
import json
import rdflib_jsonld.serializer as json_ld
import pkg_resources
import skosify

import logging
import logging.handlers

from . import __version__
from .constants import Constants
from .element import Element
from .record import InvalidRecordError, ClassificationRecord, AuthorityRecord
from .reader import MarcFileReader
from .vocabularies import Vocabularies

logging.captureWarnings(True)
warnings.simplefilter('always', DeprecationWarning)

logger = logging.getLogger()
logger.setLevel(logging.DEBUG)
formatter = logging.Formatter('[%(asctime)s %(levelname)s] %(message)s')

console_handler = logging.StreamHandler()
console_handler.setLevel(logging.DEBUG)
console_handler.setFormatter(formatter)
logger.addHandler(console_handler)


WD = Namespace('http://data.ub.uio.no/webdewey-terms#')
MADS = Namespace('http://www.loc.gov/mads/rdf/v1#')


def add_record_to_graph(graph, record, options):
    # Add record to graph

    # logger.debug('Adding: %s', record.uri)

    # Strictly, we do not need to explicitly state here that <A> and <B> are instances
    # of skos:Concept, because such statements are entailed by the definition
    # of skos:semanticRelation.
    record_uri = URIRef(record.uri)

    graph.add((record_uri, RDF.type, SKOS.Concept))

    # Add skos:topConceptOf or skos:inScheme
    for scheme_uri in record.scheme_uris:
        if record.is_top_concept:
            graph.add((record_uri, SKOS.topConceptOf, URIRef(scheme_uri)))
        else:
            graph.add((record_uri, SKOS.inScheme, URIRef(scheme_uri)))

    if record.created is not None:
        graph.add((record_uri, DCTERMS.created, Literal(record.created.strftime('%F'), datatype=XSD.date)))

    if record.modified is not None:
        graph.add((record_uri, DCTERMS.modified, Literal(record.modified.strftime('%F'), datatype=XSD.date)))

    # Add classification number as skos:notation
    if record.notation:
        if record.record_type == Constants.TABLE_RECORD:  # OBS! Sjekk add tables
            graph.add((record_uri, SKOS.notation, Literal('T' + record.notation)))
        else:
            graph.add((record_uri, SKOS.notation, Literal(record.notation)))

    # Add local control number as dcterms:identifier
    if record.control_number:
        graph.add((record_uri, DCTERMS.identifier, Literal(record.control_number)))

    # Add caption as skos:prefLabel
    if record.prefLabel:
        graph.add((record_uri, SKOS.prefLabel, Literal(record.prefLabel, lang=record.lang)))
    elif options.get('include_webdewey') and len(record.altLabel) != 0:
        # If the --webdewey flag is set, we will use the first index term as prefLabel
        caption = record.altLabel.pop(0)['term']
        if len(record.altLabel) != 0:
            caption = caption + ', …'
        graph.add((record_uri, SKOS.prefLabel, Literal(caption, lang=record.lang)))

    # Add index terms as skos:altLabel
    if options.get('include_altlabels'):
        for label in record.altLabel:
            graph.add((record_uri, SKOS.altLabel, Literal(label['term'], lang=record.lang)))

    # Add relations (SKOS:broader, SKOS:narrower, SKOS:xxxMatch, etc.)
    for relation in record.relations:
        if relation.get('uri') is not None:
            graph.add((record_uri, relation.get('relation'), URIRef(relation['uri'])))

    # Add notes
    if not options.get('exclude_notes'):
        for note in record.definition:
            graph.add((record_uri, SKOS.definition, Literal(note, lang=record.lang)))

        for note in record.note:
            graph.add((record_uri, SKOS.note, Literal(note, lang=record.lang)))

        for note in record.editorialNote:
            graph.add((record_uri, SKOS.editorialNote, Literal(note, lang=record.lang)))

        for note in record.scopeNote:
            graph.add((record_uri, SKOS.scopeNote, Literal(note, lang=record.lang)))

        for note in record.historyNote:
            graph.add((record_uri, SKOS.historyNote, Literal(note, lang=record.lang)))

        for note in record.changeNote:
            graph.add((record_uri, SKOS.changeNote, Literal(note, lang=record.lang)))

        for note in record.example:
            graph.add((record_uri, SKOS.example, Literal(note, lang=record.lang)))

    # Deprecated?
    if record.deprecated:
        graph.add((record_uri, OWL.deprecated, Literal(True)))

    # Add synthesized number components
    if options.get('include_components') and len(record.components) != 0:
        component = record.components.pop(0)
        component_uri = URIRef(record.scheme.uri('concept', collection='class', object=component))
        b1 = BNode()
        graph.add((record_uri, MADS.componentList, b1))
        graph.add((b1, RDF.first, component_uri))

        for component in record.components:
            component_uri = URIRef(record.scheme.uri('concept', collection='class', object=component))
            b2 = BNode()
            graph.add((b1, RDF.rest, b2))
            graph.add((b2, RDF.first, component_uri))
            b1 = b2

        graph.add((b1, RDF.rest, RDF.nil))

    # Add webDewey extras
    if options.get('include_webdewey'):
        for key, values in record.webDeweyExtras.items():
            for value in values:
                graph.add((record_uri, WD[key], Literal(value, lang=record.lang)))


def process_record(graph, rec, **kwargs):
    """Convert a single MARC21 classification or authority record to RDF."""
    el = Element(rec)
    leader = el.text('mx:leader')
    if leader is None:
        raise InvalidRecordError('Record does not have a leader',
                                 control_number=el.text('mx:controlfield[@tag="001"]'))
    if leader[6] == 'w':
        if kwargs.get('skip_classification'):
            return
        rec = ClassificationRecord(el, kwargs)
    elif leader[6] == 'z':
        if kwargs.get('skip_authority'):
            return
        rec = AuthorityRecord(el, kwargs)
    else:
        raise InvalidRecordError('Record is not a Marc21 Classification or Authority record',
                                 control_number=el.text('mx:controlfield[@tag="001"]'))

    if rec.is_public():
        add_record_to_graph(graph, rec, kwargs)


def process_records(records, graph=None, **options):
    n = 0
    if graph is None:
        graph = Graph()
    for record in records:
        n += 1
        try:
            process_record(graph, record, **options)
        except InvalidRecordError as e:
            record_id = e.control_number or '#%d' % n
            logger.warning('Ignoring record %s: %s', record_id, e)

    if options.get('expand'):
        logger.info('Expanding RDF via basic SKOS inference')
        skosify.infer.skos_related(graph)
        skosify.infer.skos_topConcept(graph)
        skosify.infer.skos_hierarchical(graph, narrower=True)

    if options.get('skosify'):
        logger.info('Running Skosify with config file %s', options['skosify'])
        config = skosify.config(options['skosify'])
        graph = skosify.skosify(graph, **config)

    return graph


def main():

    parser = argparse.ArgumentParser(description='Convert MARC21 Classification to SKOS/RDF')
    parser.add_argument('infile', nargs='?', help='Input XML file')
    parser.add_argument('outfile', nargs='?', help='Output RDF file')
    parser.add_argument('--version', action='version', version='%(prog)s ' + __version__)

    parser.add_argument('-v', '--verbose', dest='verbose', action='store_true', help='More verbose output')
    parser.add_argument('-o', '--outformat', dest='outformat', metavar='FORMAT', nargs='?',
                        help='Output format: turtle (default), jskos, or ndjson')

    parser.add_argument('--include', action='append', dest='include', default=[],
                        help='RDF file(s) to include in the output (e.g. to define a concept scheme). '
                             'Must be the same format as {outformat}.')

    parser.add_argument('--uri', dest='base_uri', help='Concept URI template. See vocabularies.yml for examples.')
    parser.add_argument('--scheme', dest='scheme',
                        help='Concept scheme, either an URI or a key from vocabularies.yml (For a list: mc2skos --list).')

    parser.add_argument('--whitespace', dest='whitespace', metavar='STRING',
                        help='Replace whitespaces in URI templates with this.')

    parser.add_argument('--altlabels', '--indexterms', dest='altlabels', action='store_true',
                        help='Include altlabels (from 7XX or 4XX).')
    parser.add_argument('--notes', dest='notes', action='store_true',
                        help='Include note fields (DEPRECATED as including notes is now the default).')
    parser.add_argument('--exclude_notes', dest='exclude_notes', action='store_true',
                        help='Exclude note fields.')
    parser.add_argument('--components', dest='components', action='store_true',
                        help='Include component information from 765.')
    parser.add_argument('--webdewey', dest='webdewey', action='store_true',
                        help='Include non-standard WebDewey notes from 680.')
    parser.add_argument('--skip-classification', dest='skip_classification', action='store_true',
                        help='Skip classification records')
    parser.add_argument('--skip-authority', dest='skip_authority', action='store_true',
                        help='Skip authority records')
    parser.add_argument('--expand', dest='expand', action='store_true',
                        help='Use Skosify to infer skos:hasTopConcept, skos:narrower and skos:related')
    parser.add_argument('--skosify', dest='skosify',
                        help='Run Skosify with given configuration file')

    parser.add_argument('-l', '--list-schemes', dest='list_schemes', action='store_true',
                        help='List default concept schemes.')

    args = parser.parse_args()

    if args.notes:
        warnings.warn('--notes is deprecated as including notes is now the default. '
                      'The inverse option --exclude_notes has been added to exclude notes.',
                      DeprecationWarning)

    with pkg_resources.resource_stream(__name__, 'vocabularies.yml') as fp:
        vocabularies = Vocabularies()
        vocabularies.load_yaml(fp)

    vocabularies.set_default_scheme(
        generic=args.base_uri,
        scheme=args.scheme,
        whitespace=args.whitespace
    )

    if args.list_schemes:
        print('Schemes:')
        for voc in vocabularies:
            print('- %s' % voc)
        return

    supported_formats = ['turtle', 'jskos', 'ndjson']
    if not args.outformat and args.outfile:
        ext = args.outfile.rpartition('.')[-1]
        if ext in supported_formats:
            args.outformat = ext
    if not args.outformat:
        args.outformat = 'turtle'
    elif args.outformat not in supported_formats:
        raise ValueError("Format not supported, must be one of '%s'." % "', '".join(supported_formats))

    graph = Graph()
    for filename in args.include:
        if args.outformat == 'turtle':
            graph.load(filename, format='turtle')
        else:
            graph.load(filename, format='json-ld')

    nm = graph.namespace_manager
    nm.bind('dcterms', DCTERMS)
    nm.bind('skos', SKOS)
    nm.bind('wd', WD)
    nm.bind('mads', MADS)
    nm.bind('owl', OWL)

    if args.verbose:
        console_handler.setLevel(logging.DEBUG)
    else:
        console_handler.setLevel(logging.INFO)

    if args.infile is None:
        raise ValueError('Filename not specified')

    options = {
        'include_altlabels': args.altlabels,
        'exclude_notes': args.exclude_notes,
        'include_components': args.components,
        'include_webdewey': args.webdewey,
        'skip_classification': args.skip_classification,
        'skip_authority': args.skip_authority,
        'expand': args.expand,
        'skosify': args.skosify,
        'vocabularies': vocabularies
    }

    marc = MarcFileReader(args.infile)
    graph = process_records(marc.records(), graph, **options)

    if not graph:
        logger.warning('RDF result is empty!')
        return

    if args.outfile and args.outfile != '-':
        out_file = open(args.outfile, 'wb')
    else:
        if (sys.version_info > (3, 0)):
            out_file = sys.stdout.buffer
        else:
            out_file = sys.stdout

    if args.outformat == 'turtle':
        # @TODO: Perhaps use OrderedTurtleSerializer if available, but fallback to default Turtle serializer if not?
        serializer = OrderedTurtleSerializer(graph)

        serializer.class_order = [
            SKOS.ConceptScheme,
            SKOS.Concept,
        ]
        serializer.sorters = [
            (r'/([0-9A-Z\-]+)--([0-9.\-;:]+)/e', lambda x: 'C{}--{}'.format(x[0], x[1])),  # table numbers
            (r'/([0-9.\-;:]+)/e', lambda x: 'B' + x[0]),  # standard schedule numbers
            (r'^(.+)$', lambda x: 'A' + x[0]),  # fallback
        ]

        serializer.serialize(out_file)

    elif args.outformat in ['jskos', 'ndjson']:
        s = pkg_resources.resource_string(__name__, 'jskos-context.json').decode('utf-8')
        context = json.loads(s)
        jskos = json_ld.from_rdf(graph, context)
        if args.outformat == 'jskos':
            jskos['@context'] = u'https://gbv.github.io/jskos/context.json'
            out_file.write(json.dumps(jskos, sort_keys=True, indent=2).encode('utf-8'))
        else:
            for record in jskos['@graph'] if '@graph' in jskos else [jskos]:
                record['@context'] = u'https://gbv.github.io/jskos/context.json'
                out_file.write(json.dumps(record, sort_keys=True).encode('utf-8') + b'\n')

    if out_file != sys.stdout:
        logger.info('Wrote %s: %s' % (args.outformat, args.outfile))



# encoding=utf8

from datetime import datetime
import logging
from iso639 import languages
from rdflib import URIRef
from rdflib.namespace import SKOS

from .constants import Constants
from .element import Element
from .error import InvalidRecordError, UnknownSchemeError
from .util import is_uri

logger = logging.getLogger(__name__)


class DuplicateFilter(object):
    def __init__(self):
        self.msgs = set()

    def filter(self, record):
        rv = record.msg not in self.msgs
        self.msgs.add(record.msg)
        return rv


logger.addFilter(DuplicateFilter())


class Record(object):

    def __init__(self, record, options=None):
        options = options or {}
        if isinstance(record, Element):
            self.record = record
        else:
            self.record = Element(record)

        self.control_number = None
        self.control_number_identifier = None
        self.created = None
        self.modified = None
        self.lang = None
        self.prefLabel = None
        self.altLabel = []
        self.definition = []
        self.editorialNote = []
        self.note = []
        self.components = []
        self.scopeNote = []
        self.historyNote = []
        self.changeNote = []
        self.example = []
        self.relations = []
        self.webDeweyExtras = {}
        self.deprecated = False
        self.is_top_concept = False
        self.notation = None

        self.vocabularies = options['vocabularies']
        try:
            self.scheme = self.vocabularies.get_from_record(self)
        except UnknownSchemeError as e:
            e.control_number = self.record.text('mx:controlfield[@tag="001"]')
            raise

        self.uri = None  # Concept URI
        self.scheme_uris = []  # Concept scheme URI

        self.parse(options or {})

    def get_terms(self, base='1'):
        # X00 - Personal Name
        # X10 - Corporate Name
        # X11 - Meeting Name
        # X30 - Uniform Title
        # X47 - Named Event
        # X48 - Chronological
        # X50 - Topical
        # X51 - Geographic Name
        # X53 - Uncontrolled
        # X55 - Genre/Form Term
        # X62 - Medium of Performance Term
        tags = ['@tag="%s%s"' % (base, tag) for tag in ['00', '10', '11', '30', '47', '48', '50', '51', '53', '55', '62']]

        terms = []
        for entry in self.record.all('mx:datafield[%s]' % ' or '.join(tags)):

            def reducer(value, element):
                prefix = ' '
                suffix = ''

                if value == '':
                    prefix = ''
                elif element.get('code') == 'd' and value[-1] not in [',', ';']:
                    prefix = ' ('
                    suffix = ')'
                elif element.get('code') in ['x', 'y', 'z', 'v']:
                    prefix = '--'

                return value + prefix + element.text() + suffix

            label = entry.reduce(reducer, ['a', 'd', 'x', 'y', 'z', 'v'])

            # codes = ['@code="%s"' % code for code in ['a', 'd', 'x', 'y', 'z', 'v']]
            # term_parts = entry.text('mx:subfield[%s]' % ' or '.join(codes), True)
            cn = entry.text('mx:subfield[@code="0"]')
            cni = None
            if cn is not None:
                cn = cn.split(')')
                if len(cn) == 2:
                    cn = cn[1]
                    cni = cn[0].lstrip('(')
                else:
                    cn = cn[0]
            term = {
                'value': label,
                'node': entry,
                'control_number': cn,
                'control_number_identifier': cni,
            }
            if 'isCaption' in entry.get_ess_codes():
                terms.insert(0, term)
            else:
                terms.append(term)

        return terms

    def parse(self, options):

        # 001
        self.control_number = self.record.text('mx:controlfield[@tag="001"]')

        # 010 : If present, it takes precedence over 001.
        # <https://github.com/scriptotek/mc2skos/issues/42>
        value = self.record.text('mx:datafield[@tag="010"]/mx:subfield[@code="a"]')
        if value is not None:
            self.control_number = value

        # 016 : If present, it takes precedence over 001
        # <https://github.com/scriptotek/mc2skos/issues/42>
        value = self.record.text('mx:datafield[@tag="016"]/mx:subfield[@code="a"]')
        if value is not None:
            self.control_number = value

        # 003
        self.control_number_identifier = self.record.text('mx:controlfield[@tag="003"]')

        # 005
        value = self.record.text('mx:controlfield[@tag="005"]')
        if value is not None:
            try:
                self.modified = datetime.strptime(value, '%Y%m%d%H%M%S.%f')
            except ValueError:
                logger.warning('Record %s: Ignoring invalid date in 005 field: %s', self.control_number, value)

        # 040: Record Source
        lang = self.record.text('mx:datafield[@tag="040"]/mx:subfield[@code="b"]') or 'eng'
        self.lang = languages.get(part2b=lang).part1

    def is_public(self):
        return True

    def append_relation(self, scheme_code, scheme_type, relation, **kwargs):
        try:
            scheme = self.vocabularies.get(scheme_code, edition=kwargs.get('edition'))
            uri = scheme.uri('concept', **kwargs)
        except UnknownSchemeError:
            tag = ' in field %s' % kwargs.get('tag') if kwargs.get('tag') else ''
            logger.warning((
                'Found links to "%s"%s, but mc2skos doesn\'t know the URI pattern of'
                ' that vocabulary, so no SKOS mappings were generated. See'
                ' <https://github.com/scriptotek/mc2skos#uris> for more info.'
            ) % (scheme_code, tag))
            return

        if uri:
            self.relations.append({
                'uri': uri,
                'relation': relation,
            })

    def get_mappings(self):
        # Get a list of possible mappings.

        for field in self.record.all('mx:datafield[@tag="024"]'):
            control_number = field.text('mx:subfield[@code="a"]')
            scheme_code = field.text('mx:subfield[@code="2"]')
            if scheme_code != 'uri':
                yield {
                    'scheme_code': scheme_code,
                    'relation': SKOS.exactMatch,
                    'control_number': control_number,
                    'tag': '024',
                }

        for heading in self.get_terms('7'):
            relation = None
            for sf in heading['node'].all('mx:subfield'):
                if sf.get('code') == '4':
                    if is_uri(sf.text()):
                        relation = URIRef(sf.text())
                    else:
                        relation = {
                            '=EQ': SKOS.exactMatch,
                            '~EQ': SKOS.closeMatch,
                            'BM': SKOS.broadMatch,
                            'NM': SKOS.narrowMatch,
                            'RM': SKOS.relatedMatch,
                        }.get(sf.text())  # None if no match

                elif sf.get('code') == '0' or sf.get('code') == '1':
                    # Note: Default value might change in the future
                    relation = relation if relation else SKOS.closeMatch

                    if is_uri(sf.text()):
                        self.relations.append({
                            'uri': sf.text(),
                            'relation': relation,
                        })
                    else:
                        scheme_code = {
                            '0': 'a',  # Library of Congress Subject Headings
                            '1': 'b',  # LC subject headings for children's literature
                            '2': 'c',  # Medical Subject Headings
                            '3': 'd',  # National Agricultural Library subject authority file
                            '4': 'n',  # Source not specified
                            '5': 'k',  # Canadian Subject Headings
                            '6': 'v',  # Répertoire de vedettes-matière
                            '7': heading['node'].text('mx:subfield[@code="2"]'),  # Source specified in subfield $2
                        }.get(heading['node'].get('ind2'))

                        yield {
                            'scheme_code': scheme_code,
                            'relation': relation,
                            'control_number': sf.text(),
                            'tag': heading['node'].get('tag'),
                        }


class ClassificationRecord(Record):

    def __init__(self, record, options=None):
        options = options or {}

        super(ClassificationRecord, self).__init__(record, options)

    def generate_uris(self):
        # Generate URIs from scheme
        self.scheme_uris = []

        if self.record_type == Constants.TABLE_RECORD:
            table = self.table if self.table is not None else ''
            uri = self.scheme.uri('scheme', collection='table', object=table)
            if uri:
                self.scheme_uris.append(uri)

        obj = 'edition' if self.scheme.edition is not None else ''
        uri = self.scheme.uri('scheme', collection='scheme', object=obj)
        if uri:
            self.scheme_uris.append(uri)

        # Record URI
        self.uri = self.scheme.uri('concept', collection='class', object=self.notation, control_number=self.control_number)

    def parse(self, options):

        super(ClassificationRecord, self).parse(options)

        # 008
        value = self.record.text('mx:controlfield[@tag="008"]')
        self.created, self.record_type, self.number_type, self.display, self.synthesized, self.deprecated = self.parse_008(value)

        # 153: Classification number
        element = self.record.first('mx:datafield[@tag="153"]')
        if element is None:
            raise InvalidRecordError('153 field is missing', control_number=self.control_number)
        self.table, self.notation, self.is_top_concept, parent_notation, self.prefLabel = self.parse_153(element)

        if self.record_type is None:
            logger.warning('Record does not have a 008 field, will try to guess type.')
            if self.table is None:
                self.record_type = Constants.SCHEDULE_RECORD
            else:
                self.record_type = Constants.TABLE_RECORD

        # Now we have enough information to generate URIs
        self.generate_uris()
        if parent_notation is not None:
            parent_uri = self.scheme.uri('concept', collection='class', object=parent_notation)
            if parent_uri is not None:
                self.relations.append({
                    'uri': parent_uri,
                    'relation': SKOS.broader
                })

        # 253 : Complex See Reference (R)
        # Example:
        # <mx:datafield tag="253" ind1="2" ind2=" ">
        #   <mx:subfield code="i">Klassifiser</mx:subfield>
        #   <mx:subfield code="t">naturhistorie</mx:subfield>
        #   <mx:subfield code="i">i</mx:subfield>
        #   <mx:subfield code="a">508</mx:subfield>
        #   <mx:subfield code="9">ess=nce</mx:subfield>
        # </mx:datafield>
        #
        for entry in self.record.all('mx:datafield[@tag="253"]'):
            self.editorialNote.append(entry.stringify())  # Constants.COMPLEX_SEE_REFERENCE

        # 353 : Complex See Also Reference (R)
        # Example:
        # <mx:datafield tag="353" ind1=" " ind2=" ">
        #   <mx:subfield code="i">Se også</mx:subfield>
        #   <mx:subfield code="a">900</mx:subfield>
        #   <mx:subfield code="i">for en</mx:subfield>
        #   <mx:subfield code="t">bred beskrivelse av situasjon og vilkår for intellektuell virksomhet</mx:subfield>
        #   <mx:subfield code="9">ess=nsa</mx:subfield>
        # </mx:datafield>
        for entry in self.record.all('mx:datafield[@tag="353"]'):
            self.editorialNote.append(entry.stringify())  # Constants.COMPLEX_SEE_ALSO_REFERENCE

        # 680 : Scope note
        # Example:
        # <mx:datafield tag="680" ind1="1" ind2=" ">
        #   <mx:subfield code="i">Her:</mx:subfield>
        #   <mx:subfield code="t">Addisjon</mx:subfield>
        #   <mx:subfield code="i">,</mx:subfield>
        #   <mx:subfield code="t">subtraksjon</mx:subfield>
        #   <mx:subfield code="i">,</mx:subfield>
        #   <mx:subfield code="t">multiplikasjon</mx:subfield>
        #   <mx:subfield code="i">,</mx:subfield>
        #   <mx:subfield code="t">divisjon</mx:subfield>
        #   <mx:subfield code="9">ess=nch</mx:subfield>
        # </mx:datafield>
        #
        for entry in self.record.all('mx:datafield[@tag="680"]'):
            ess = entry.get_ess_codes()
            if 'ndf' in ess:
                self.definition.append(entry.stringify())  # Constants.DEFINITION
            else:
                self.scopeNote.append(entry.stringify())  # Constants.SCOPE_NOTE
                topics = [t.capitalize() for t in entry.text('mx:subfield[@code="t"]', True)]
                for topic in topics:
                    if 'nvn' in ess:
                        self.webDeweyExtras['variantName'] = self.webDeweyExtras.get('variantName', []) + [topic]
                    elif 'nch' in ess:
                        self.webDeweyExtras['classHere'] = self.webDeweyExtras.get('classHere', []) + [topic]
                    elif 'nin' in ess:
                        self.webDeweyExtras['including'] = self.webDeweyExtras.get('including', []) + [topic]
                    elif 'nph' in ess:
                        self.webDeweyExtras['formerName'] = self.webDeweyExtras.get('formerName', []) + [topic]

        # 683 : Application Instruction Note
        # Example:
        # <mx:datafield tag="683" ind1="1" ind2=" ">
        #   <mx:subfield code="i">Ordnes alfabetisk etter</mx:subfield>
        #   <mx:subfield code="t">navnet på datamaskinen eller prosessoren</mx:subfield>
        #   <mx:subfield code="i">, f.eks.</mx:subfield>
        #   <mx:subfield code="t">IBM System z9®</mx:subfield>
        #   <mx:subfield code="9">ess=nal</mx:subfield>
        # </mx:datafield>
        #
        for entry in self.record.all('mx:datafield[@tag="683"]'):
            self.editorialNote.append(entry.stringify())  # Constants.APPLICATION_INSTRUCTION_NOTE

        # 685 : History note
        # Example:
        #  <mx:datafield tag="685" ind2="0" ind1="1">
        #    <mx:subfield code="i">Klassifiseres nå i</mx:subfield>
        #    <mx:subfield code="a">512.901</mx:subfield>
        #    <mx:subfield code="c">512.909</mx:subfield>
        #    <mx:subfield code="9">ess=nrl</mx:subfield>
        #  </mx:datafield>
        #
        for entry in self.record.all('mx:datafield[@tag="685"]'):
            self.historyNote.append(entry.stringify())  # Constants.HISTORY_NOTE

        # 684 : Auxiliary Instruction Note
        # 694 : ??? Note : Non-standard code for 684 'Auxiliary Instruction Note' ??
        # Example:
        #   <mx:datafield tag="694" ind2=" " ind1=" ">
        #     <mx:subfield code="i">De fleste verker om seletøy og tilbehør klassifiseres med hester i</mx:subfield>
        #     <mx:subfield code="a">636.10837</mx:subfield>
        #     <mx:subfield code="9">ess=nml</mx:subfield>
        #   </mx:datafield>
        #
        for entry in self.record.all('mx:datafield[@tag="684" or @tag="694"]'):
            self.editorialNote.append(entry.stringify())

        # 7XX Index terms
        for heading in self.get_terms('7'):
            self.altLabel.append({
                'term': heading['value']
            })

        # 7XX: Heading Linking Entries
        for mapping in self.get_mappings():
            self.append_relation(
                mapping['scheme_code'],
                None,
                mapping['relation'],
                control_number=mapping['control_number'],
                tag=mapping['tag']
            )

        # 765 : Synthesized Number Components
        for entry in reversed(list(self.record.all('mx:datafield[@tag="765"]'))):

            if entry.text('mx:subfield[@code="u"]') is None:
                logger.debug('Built number without components specified: %s', self.notation)

            table = ''
            rootno = ''
            for sf in entry.all('mx:subfield'):
                if sf.get('code') == 'b':    # Base number
                    if len(self.components) == 0:
                        self.components.append(table + sf.text())
                        table = ''
                elif sf.get('code') == 'r':    # Root number
                    rootno = sf.text()
                elif sf.get('code') == 'z':    # Table identification
                    table = '{0}--'.format(sf.text())
                # elif sf.get('code') == 't':    # Digits added from internal subarrangement or add table
                #     self.components.append(sf.text())
                elif sf.get('code') == 's':  # Digits added from classification number in schedule or external table
                    if sf.text() is None:
                        logger.warning('Class %s has blank 765 $s subfield. This should be fixed.', self.notation)
                    else:
                        tmp = rootno + sf.text()
                        if len(tmp) > 3:
                            tmp = tmp[:3] + '.' + tmp[3:]
                        self.components.append(table + tmp)
                        table = ''
                # elif sf.get('code') not in ['9', 'u']:
                #     print sf.get('code'), sf.text, class_no

    @staticmethod
    def parse_008(value):
        # Parse the 008 field text

        if value is None:
            return None, None, None, True, False, False

        created = datetime.strptime(value[:6], '%y%m%d')

        if value[6] == 'a':
            record_type = Constants.SCHEDULE_RECORD
        elif value[6] == 'b':
            record_type = Constants.TABLE_RECORD
        elif value[6] == 'e':
            record_type = Constants.EXTERNAL_SUMMARY
        elif value[6] == 'i':
            record_type = Constants.INTERNAL_SUMMARY_OF_SCHEDULE_NUMBER
        elif value[6] == 'j':
            record_type = Constants.INTERNAL_SUMMARY_OF_TABLE_NUMBER
        elif value[6] == 'm':
            record_type = Constants.MANUAL_NOTE_RECORD
        elif value[6] == '1':  # @TODO: Find out what this means! It's not documented
            record_type = Constants.SCHEDULE_RECORD
        else:
            logger.warning('Unknown value in 008/6: %s', value[6])
            record_type = Constants.UNKNOWN

        if value[7] == 'a':
            number_type = Constants.SINGLE_NUMBER
        elif value[7] == 'b':
            number_type = Constants.NUMBER_SPAN
        elif value[7] == 'c':
            number_type = Constants.SUMMARY_NUMBER_SPAN
        else:
            number_type = Constants.UNKNOWN

        deprecated = False
        if value[8] == 'd' or value[8] == 'e':
            deprecated = True

        if value[12] == 'b':
            synthesized = True
        else:
            synthesized = False

        if value[13] == 'a':
            # Displayed in standard schedules or tables
            display = True
        elif value[13] == 'b':
            # Extended display
            # These records show up in search in the WebDewey interface
            display = True
        elif value[13] == 'h':
            # Historical information, not intended for display.
            # These records do not show up in search in the WebDewey interface
            display = False
        elif value[7] == 'n':
            # Other information, not intended for display
            # These records do not show up in search in the WebDewey interface
            display = False
        else:
            logger.warning('Unknown value in 008/13: %s', value[13])
            display = False

        return created, record_type, number_type, display, synthesized, deprecated

    @staticmethod
    def parse_153(element):
        # Parse the 153 field

        table = None
        add_table = None
        notation = None
        parent_notation = None
        buf = ''
        caption = None  # Note: Synthesized classes do not have captions, that's ok
        zbuf = ''
        is_top_concept = True
        parts = []

        buf = [{'code': sf.get('code'), 'value': sf.text()} for sf in element.all('mx:subfield')]

        mode = 'notation'
        for idx, subfield in enumerate(buf):
            if subfield['code'] == 'z':
                table = subfield['value']

            elif subfield['code'] == 'y':
                add_table = subfield['value']

            elif subfield['code'] == 'a' and mode == 'notation':
                if add_table == '1':
                    notation += ':'
                elif add_table is not None:
                    notation += ':%s;' % add_table
                elif table is not None:
                    notation = '%s--' % table
                else:
                    notation = ''
                notation += subfield['value']
                add_table = None

            elif subfield['code'] == 'c' and mode == 'notation':
                notation += '-' + subfield['value']

            elif subfield['code'] == 'e' and mode in ['notation', 'parent']:
                parent_notation = ''
                if add_table == '1':
                    parent_notation += ':'
                elif add_table is not None:
                    parent_notation += ';%s:' % add_table
                elif table is not None:
                    parent_notation = '%s--' % table
                parent_notation += subfield['value']
                add_table = None
                mode = 'parent'

            elif subfield['code'] == 'f' and mode == 'parent':
                parent_notation += '-' + subfield['value']

            elif subfield['code'] == 'j':
                caption = subfield['value']

            elif subfield['code'] == 'h':
                # In the ddc21 examples, the parent class numbers (153 $e, $f) are not included,
                # but the parent class headings are (153 $h). We do not make any attempt of mapping
                # the headings to the classes, but just take note that this is not a top concept.
                is_top_concept = False

            else:
                mode = 'other'

        if parent_notation is not None:
            is_top_concept = False

        return table, notation, is_top_concept, parent_notation, caption

    def is_public(self):
        if not self.display:
            # This is a record not displayed in standard schedules or tables
            # or in extended display. It could be a deleted (not deprecated)
            # class.
            logger.debug('%s is not intended for display', self.notation)
            return False

        if self.record_type not in [Constants.SCHEDULE_RECORD, Constants.TABLE_RECORD]:
            logger.debug('%s is a type %s', self.notation, self.record_type)
            return False

        include_add_table_numbers = False  # @TODO: Make argparse option
        if self.notation.find(':') != -1 and not include_add_table_numbers:
            logger.debug('%s is an add table number', self.notation)
            return False

        return True


class AuthorityRecord(Record):

    def __init__(self, record, options=None):
        super(AuthorityRecord, self).__init__(record, options)

    def generate_uris(self):
        # Generate URIs from scheme
        self.scheme_uris = []

        scheme_uri = self.scheme.uri('scheme')
        if scheme_uri:
            self.scheme_uris.append(scheme_uri)

        # Record URI
        self.uri = self.scheme.uri('concept', control_number=self.control_number)

    @staticmethod
    def get_class_number(el):
        number_start = el.text('mx:subfield[@code="a"]')
        number_end = el.text('mx:subfield[@code="b"]')
        if number_end is not None:
            return '{}-{}'.format(number_start, number_end)
        else:
            return number_start

    def parse(self, options):
        super(AuthorityRecord, self).parse(options)

        # Now we have enough information to generate URIs
        self.generate_uris()

        leader = self.record.text('mx:leader')
        if leader[5] in ['d', 'o', 's', 'x']:
            self.deprecated = True

        # 008
        field_008 = self.record.text('mx:controlfield[@tag="008"]')
        if field_008:
            self.created = datetime.strptime(field_008[:6], '%y%m%d')

        # 065: Other Classification Number
        el = self.record.first('mx:datafield[@tag="065"]')
        if el is not None:
            self.append_relation(
                el.text('mx:subfield[@code="2"]'),
                ClassificationRecord,
                SKOS.exactMatch,
                object=self.get_class_number(el),
                tag='065'
            )

        # 080: Universal Decimal Classification Number
        el = self.record.first('mx:datafield[@tag="080"]')
        if el is not None:
            self.append_relation(
                'udc',
                ClassificationRecord,
                SKOS.exactMatch,
                object=self.get_class_number(el),
                tag='080'
            )

        # 083: Dewey Decimal Classification Number
        el = self.record.first('mx:datafield[@tag="083"]')
        if el is not None:
            self.append_relation(
                'ddc',
                ClassificationRecord,
                SKOS.exactMatch,
                collection='class',
                object=self.get_class_number(el),
                edition=el.text('mx:subfield[@code="2"]'),
                tag='083'
            )

        # 1XX Heading
        for heading in self.get_terms('1'):
            self.prefLabel = heading['value']

        # 4XX: See From Tracings
        for heading in self.get_terms('4'):
            self.altLabel.append({
                'term': heading['value']
            })

        # 5XX: See Also From Tracings
        for heading in self.get_terms('5'):
            local_id = heading['node'].text('mx:subfield[@code="0"]')
            if local_id:
                if local_id:
                    sf_w = heading['node'].text('mx:subfield[@code="w"]')
                    sf_4 = heading['node'].text('mx:subfield[@code="4"]')

                    if sf_w == 'g':
                        relation = SKOS.broader
                    elif sf_w == 'h':
                        relation = SKOS.narrower
                    elif sf_w == 'r' and sf_4 is not None and is_uri(sf_4):
                        relation = URIRef(sf_4)
                    else:
                        relation = SKOS.related

                    if is_uri(local_id):
                        self.relations.append({
                            'uri': local_id,
                            'relation': relation,
                        })
                    else:
                        self.append_relation(
                            self.scheme.code,
                            self.scheme.type,
                            relation,
                            control_number=local_id,
                            tag=heading['node'].get('tag')
                        )

        # 667 : Nonpublic General Note
        # madsrdf:editorialNote
        for entry in self.record.all('mx:datafield[@tag="667"]'):
            self.editorialNote.append(entry.stringify(subfields=['a']))

        # 670 : Source Data Found
        # Citation for a consulted source in which information is found related in some
        # manner to the entity represented by the authority record or related entities.
        for entry in self.record.all('mx:datafield[@tag="670"]'):
            self.note.append('Source: ' + entry.stringify(subfields=['a']))

        # 677 : Definition
        for entry in self.record.all('mx:datafield[@tag="677"]'):
            self.definition.append(entry.stringify(subfields=['a']))

        # 678 : Biographical or Historical Data
        # Summary of the essential biographical, historical, or other information about the 1XX heading
        # madsrdf:note
        for entry in self.record.all('mx:datafield[@tag="678"]'):
            self.note.append(entry.stringify(subfields=['a', 'b']))

        # 680 : Public General Note
        # madsrdf:note
        for entry in self.record.all('mx:datafield[@tag="680"]'):
            self.note.append(entry.stringify(subfields=['a', 'i']))

        # 681 : Subject Example Tracing Note
        # madsrdf:exampleNote
        for entry in self.record.all('mx:datafield[@tag="681"]'):
            self.example.append(entry.stringify(subfields=['a', 'i']))

        # 682 : Deleted Heading Information
        # Explanation for the deletion of an established heading or subdivision record from an authority file.
        # madsrdf:changeNote
        for entry in self.record.all('mx:datafield[@tag="682"]'):
            self.changeNote.append(entry.stringify(subfields=['a', 'i']))

        # 688 : Application History Note
        # Information that documents changes in the application of a 1XX heading.
        # madsrdf:historyNote
        for entry in self.record.all('mx:datafield[@tag="688"]'):
            self.historyNote.append(entry.stringify(subfields=['a']))

        # 7XX: Heading Linking Entries
        for mapping in self.get_mappings():
            self.append_relation(
                mapping['scheme_code'],
                None,
                mapping['relation'],
                control_number=mapping['control_number'],
                tag=mapping['tag']
            )



# encoding=utf-8
import unittest
import pytest
import os
import sys
import glob
import re
from lxml import etree
from mc2skos.reader import MarcFileReader
from mc2skos.mc2skos import process_records
from mc2skos.vocabularies import Vocabularies
from rdflib.namespace import RDF, SKOS, OWL, DCTERMS, Namespace
from rdflib import URIRef, Literal, Graph


with open('mc2skos/vocabularies.yml') as fp:
    vocabularies = Vocabularies()
    vocabularies.load_yaml(fp)


def examples(pattern):
    pattern = '^(examples/%s)\.xml$' % (pattern)
    files = glob.glob('examples/*.xml')
    files = filter(lambda x: re.match(pattern, x), files)

    return [(MarcFileReader(f), re.match(pattern, f)) for f in files]


def check_processing(marc, expect, **kwargs):
    kwargs['vocabularies'] = vocabularies
    graph = process_records(marc.records(), **kwargs)
    filename = re.sub('xml$', 'ttl', marc.name)

    graph.namespace_manager.bind('skos', SKOS)
    graph.namespace_manager.bind('owl', OWL)
    graph.namespace_manager.bind('dcterms', DCTERMS)

    if os.path.isfile(filename):
        expect.parse(filename, format='turtle')
    elif len(graph) > 0:
        graph.serialize(destination=filename, format='turtle')

    # graph.serialize(destination=sys.stdout, format='turtle')

    for triple in expect:
        assert triple in graph


@pytest.mark.parametrize('marc,match',
                         examples('ddc(?P<edition>\d{2})(?P<lang>[a-z]+)-'
                                  '(?P<notation>((?P<table>\d+)--)?[\d.]+-?[\d.]*)'))
def test_ddc_example(marc, match):
    edition = match.group('edition')
    notation = match.group('notation')
    table = match.group('table')

    expect = Graph()
    uri = URIRef(u'http://dewey.info/class/' + notation + '/e' + edition + '/')
    expect.add((uri, RDF.type, SKOS.Concept))
    if table:
        notation = "T" + notation
    expect.add((uri, SKOS.notation, Literal(notation)))

    check_processing(marc, expect, include_webdewey=True)


@pytest.mark.parametrize('marc,match',
                         examples('(?P<scheme>bk|asb)-(?P<notation>[0-9.]+)'))
def test_bk_asb_example(marc, match):
    scheme = match.group('scheme')
    notation = match.group('notation')

    expect = Graph()
    uri = URIRef(u'http://uri.gbv.de/terminology/{}/{}'.format(scheme, notation))
    expect.add((uri, RDF.type, SKOS.Concept))

    check_processing(marc, expect, include_altlabels=True)


@pytest.mark.parametrize('marc,match', examples('rvk(-.*)?'))
def test_rvk_example(marc, match):

    options = {
        'include_altlabels': True,
    }

    check_processing(marc, Graph(), **options)

authority_vocabularies = {
    'lcgft': 'http://id.loc.gov/authorities/genreForms/',
    'lcsh': 'http://id.loc.gov/authorities/subjects/',
    'noubomn': 'http://data.ub.uio.no/realfagstermer/',
    'noubojur': 'http://data.ub.uio.no/lskjema/',
    'humord': 'http://data.ub.uio.no/humord/',
    'nalt': 'http://lod.nal.usda.gov/nalt/',
    'gnd': 'http://d-nb.info/gnd/',
}


@pytest.mark.parametrize('marc,match',
                         examples('(?P<vocabulary>' + '|'.join(authority_vocabularies.keys()) + ')-(?P<control_number>.+)'))
def test_authority_example(marc, match):
    vocabulary = match.group('vocabulary')
    control_number = match.group('control_number')

    uri_base = authority_vocabularies[vocabulary]

    expect = Graph()
    uri = URIRef(uri_base + control_number)
    expect.add((uri, RDF.type, SKOS.Concept))

    if vocabulary == 'gnd':
        # For GND, we need to specify --scheme gnd
        # See: https://github.com/scriptotek/mc2skos/issues/56
        vocabularies.set_default_scheme(scheme='gnd')

    check_processing(marc, expect, include_altlabels=True)
    vocabularies.set_default_scheme()

if __name__ == '__main__':
    unittest.main()



# encoding=utf8


class Constants(object):
    SCHEDULE_RECORD = 'schedule_record'
    TABLE_RECORD = 'table_record'
    INTERNAL_SUMMARY_OF_SCHEDULE_NUMBER = 'internal_summary_of_schedule_number'
    EXTERNAL_SUMMARY = 'external_summary'
    INTERNAL_SUMMARY_OF_TABLE_NUMBER = 'internal_summary_of_table_number'
    MANUAL_NOTE_RECORD = 'manual_note_record'

    UNKNOWN = 'unknown'

    SINGLE_NUMBER = 'single_number'
    NUMBER_SPAN = 'number_span'
    SUMMARY_NUMBER_SPAN = 'summary_number_span'

    COMPLEX_SEE_REFERENCE = 'nce'
    COMPLEX_SEE_ALSO_REFERENCE = 'nsa'
    DEFINITION = 'ndf'
    SCOPE_NOTE = 'scope_note'
    APPLICATION_INSTRUCTION_NOTE = 'application_instruction_note'
    HISTORY_NOTE = 'history_note'
    AUXILIARY_INSTRUCTION_NOTE = 'auxiliary_instruction_note'





class InvalidRecordError(RuntimeError):

    def __init__(self, msg, control_number=None):
        super(InvalidRecordError, self).__init__(msg)
        self.control_number = control_number


class UnknownSchemeError(InvalidRecordError):

    def __init__(self, code=None, **kwargs):
        if code is None:
            msg = 'Could not find classification scheme or subject vocabulary code.'
        else:
            msg = 'Cannot generate URIs for unknown classification scheme or subject vocabulary "%s".' % code
        super(UnknownSchemeError, self).__init__(msg)
        self.code = code
        self.control_number = kwargs.get('control_number')



#!/usr/bin/env python
# encoding=utf-8
from __future__ import print_function
import os
import sys

try:
    from setuptools import setup
except ImportError:
    print("This package requires 'setuptools' to be installed.")
    sys.exit(1)

here = os.path.abspath(os.path.dirname(__file__))
README = open(os.path.join(here, 'README.rst')).read()

setup(name='mc2skos',
      version='0.12.0',  # Use bumpversion to update
      description='Convert Marc21 Classification records in MARC/XML to SKOS/RDF ',
      long_description=README,
      classifiers=[
          'Programming Language :: Python',
          'Programming Language :: Python :: 2.7',
          'Programming Language :: Python :: 3.6',
          'Programming Language :: Python :: 3.7',
          'Programming Language :: Python :: 3.8',
      ],
      keywords='marc rdf skos',
      author='Dan Michael O. Heggø',
      author_email='danmichaelo@gmail.com',
      url='https://github.com/scriptotek/mc2skos',
      license='MIT',
      install_requires=['rdflib[sparql]',
                        'rdflib-jsonld',
                        'lxml',
                        'otsrdflib>=0.5.0,<0.6.0',
                        'iso-639',
                        'pyyaml',
                        'future',
                        'skosify>=2.0.1'
                        ],
      setup_requires=['rdflib', 'pytest-runner>=2.9'],
      tests_require=['pytest', 'pytest-pep8', 'pytest-cov'],
      packages=['mc2skos'],
      entry_points={'console_scripts': ['mc2skos=mc2skos.mc2skos:main']},
      package_data={
          'mc2skos': ['jskos-context.json', 'vocabularies.yml']
      },
      include_package_data=True
      )



from future.utils import python_2_unicode_compatible
import re
import yaml
from .error import UnknownSchemeError
from .record import AuthorityRecord, ClassificationRecord
from .util import is_str


@python_2_unicode_compatible
class Vocabularies(object):

    def __init__(self):
        self.entries = {}
        self.default_scheme = None

    def __iter__(self):
        for val in self.entries.values():
            yield val

    def set_default_scheme(self, generic=None, concept=None, scheme=None, whitespace=None):
        # Set URI templates manually. This will override values in vocabulary.yml.
        if generic is None and concept is None and scheme is None:
            self.default_scheme = None
            return

        options = {
            'base_uri': generic,
            'concept': concept,
            'scheme': scheme,
            'whitespace': whitespace,
        }
        if scheme in self.entries:
            self.default_scheme = self.entries[scheme]
        else:
            self.default_scheme = ConceptScheme(options=options)

    def load_yaml(self, file):
        data = yaml.safe_load(file)
        for concept_type_key, vocabs in data.items():
            concept_type = {
                'classification_schemes': ClassificationRecord,
                'subject_schemes': AuthorityRecord,
            }.get(concept_type_key)
            for scheme_code, options in vocabs.items():
                if is_str(options):
                    options = {'base_uri': options}
                self.entries[scheme_code] = ConceptScheme(concept_type, scheme_code, options=options)

    def get(self, scheme_code, edition=None):
        if scheme_code == 'n':
            raise UnknownSchemeError()
        if scheme_code not in self.entries:
            raise UnknownSchemeError(scheme_code)
        scheme = self.entries[scheme_code]
        if edition is not None:
            key = '%s-%s' % (scheme_code, edition)
            if key not in self.entries:
                self.entries[key] = scheme.with_edition(edition)
            return self.entries[key]
        return scheme

    def get_from_record(self, record):

        if self.default_scheme is not None:
            return self.default_scheme

        if isinstance(record, AuthorityRecord):
            field_008 = record.record.text('mx:controlfield[@tag="008"]')
            if field_008:
                code = field_008[11]
                if code == 'z':
                    code = record.record.text('mx:datafield[@tag="040"]/mx:subfield[@code="f"]')

                if code:
                    return self.get(code)

        if isinstance(record, ClassificationRecord):
            code = record.record.text('mx:datafield[@tag="084"]/mx:subfield[@code="a"]')
            edition = record.record.text('mx:datafield[@tag="084"]/mx:subfield[@code="c"]')
            if code:
                return self.get(code, edition=edition)

        raise UnknownSchemeError()


@python_2_unicode_compatible
class ConceptScheme(object):

    def __init__(self, concept_type=None, code=None, edition=None, options=None):
        options = options or {}
        self.type = concept_type
        self.code = code  # Can be None if URI template is specified in options
        self.edition = edition
        self.options = options
        self.edition_numeric = re.sub('[^0-9]', '', edition or '')

        self.uri_templates = {
            'concept': options.get('concept') or options.get('base_uri'),
            'scheme': options.get('scheme') or options.get('base_uri'),
        }

        self.whitespace = options.get('whitespace') or '-'

    def with_edition(self, edition):
        # Get a specific edition of this scheme
        return ConceptScheme(self.type, self.code, edition, self.options)

    def __repr__(self):
        if self.edition is not None:
            return u'%s (%s ed.)' % (self.code, self.edition)
        return u'%s' % (self.code)

    def uri(self, uri_type, **kwargs):
        if uri_type not in self.uri_templates:
            raise ValueError('Unknown URI type: %s' % uri_type)

        uri_template = self.uri_templates[uri_type]
        if uri_template is None:
            raise UnknownSchemeError(
                self.code,
                message='No URI template found for URIs of type "%s" in vocabulary "%s"' % (uri_type, self.code)
            )

        kwargs['edition'] = self.edition_numeric
        if uri_type == 'scheme':
            kwargs['control_number'] = ''

        if kwargs.get('control_number') is not None:
            # Remove organization prefix in parenthesis:
            kwargs['control_number'] = re.sub(r'^\(.+\)(.+)$', '\\1', kwargs['control_number'])

        # Process field[start:end]

        def process_formatter(matches):
            start = int(matches.group('start')) if matches.group('start') else None
            end = int(matches.group('end')) if matches.group('end') else None
            value = kwargs[matches.group('param')][start:end]
            if len(value) == 0:
                # Empty string can be used for the scheme URI.
                # Trying to convert this to decimal or float will fail!
                formatter_str = '{0}'
            else:
                formatter_str = '{0' + matches.group('formatter') + '}' if matches.group('formatter') else '{0}'
                if 'd' in formatter_str:
                    value = int(value)
                elif 'f' in formatter_str:
                    value = float(value)

            return formatter_str.format(value)

        uri_template = re.sub(
            r'\{(?P<param>[a-z_]+)(?:\[(?P<start>\d+)?:(?P<end>\d+)?\])?(?P<formatter>[:!][^\}]+)?\}',
            process_formatter,
            uri_template
        )

        uri = uri_template.format(**kwargs)

        # replace whitespaces in URI
        return uri.replace(' ', self.whitespace)



# encoding=utf8
import logging
import time
from lxml import etree

logger = logging.getLogger(__name__)


class MarcFileReader:
    """Read records from a MARC XML file."""

    def __init__(self, name):
        self.name = name

    def records(self):
        logger.info('Parsing: %s', self.name)
        n = 0
        t0 = time.time()
        record_tag = '{http://www.loc.gov/MARC21/slim}record'
        for _, record in etree.iterparse(self.name, tag=record_tag):
            yield record
            record.clear()
            n += 1
            if n % 500 == 0:
                logger.info('Read %d records (%.f recs/sec)', n, (float(n) / (time.time() - t0)))



# encoding=utf8

from functools import reduce
from lxml import etree
import re


class Element(object):

    nsmap = {
        'mx': 'http://www.loc.gov/MARC21/slim',
        'marc': 'http://www.loc.gov/MARC21/slim',
    }

    def __init__(self, data):
        if isinstance(data, etree._Element):
            self.node = data
        else:
            self.node = etree.fromstring(data)

    def get(self, name):
        return self.node.get(name)

    def all(self, xpath):
        # Yields all nodes matching the xpath
        for res in self.node.xpath(xpath, namespaces=self.nsmap):
            yield Element(res)

    def first(self, xpath):
        # Returns first node or None
        for res in self.all(xpath):
            return res

    def text(self, xpath=None, all=False):
        # xpath: the xpath
        # all: True to return an array with the text content for all matching elements.
        #      False to return a string with the text content of the first matching element, or None.
        # Returns text content of first node or None

        def flatten_text(node):
            # Captions can include Processing Instruction tags, like in this example
            # (linebreaks added):
            #
            #   <mx:subfield xmlns:mx="http://www.loc.gov/MARC21/slim"
            #                xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance" code="t">
            #     <?ddc fotag="fo:inline" font-style="italic"?>L
            #       <?ddc fotag="fo:inline" vertical-align="super" font-size="70%"?>p
            #       <?ddc fotag="/fo:inline"?>
            #     <?ddc fotag="/fo:inline"?>-rom
            #   </mx:subfield>
            #
            # The code below just strips away the PI tags, giving "Lp-rom" for this example.
            children = node.getchildren()
            if len(children) != 0:
                value = ''
                for child in children:
                    if child.tail is not None:
                        value += child.tail
            else:
                value = node.text
            return value

        if xpath is None:
            return flatten_text(self.node)
        if all:
            return [flatten_text(res.node) for res in self.all(xpath) if res.node.text is not None]
        for res in self.all(xpath):
            return flatten_text(res.node)  # return text of first element

    def get_ess_codes(self):
        return [x[4:] for x in self.node.xpath('mx:subfield[@code="9"]/text()', namespaces=self.nsmap) if x.find('ess=') == 0]

    def reduce(self, fn, subfields=['a', 'c', 'i', 't', 'x'], initializer=''):
        codes = ['@code="%s"' % code for code in subfields]
        return reduce(fn, self.all('mx:subfield[%s]' % ' or '.join(codes)), initializer)

    def stringify(self, subfields=['a', 'c', 'i', 't', 'x']):
        def inner(label, subfield):
            code = subfield.get('code')
            value = subfield.text()
            if value is None:
                return label

            # Check if we need to add a separator
            if code == 'c':
                # Treat $c as the end of a number span, which is correct for the 6XX fields
                # in MARC21 Classification. In Marc21 Authority, $c generally seems to be
                # undefined, but we might add some checks here if there are some $c subfields
                # that need to be treated differently.
                value = '-' + value

            elif len(label) != 0 and not re.match(r'[.\?#@+,<>%~`!$^&\(\):;\]]', value[0]):
                # Unless the subfield starts with a punctuation character, we will add a space.
                value = ' ' + value

            return label + value

        return self.reduce(inner, subfields)
